﻿namespace UsefulTools.Wpf
{
    using System;
    using System.Collections.Generic;
    using System.Reflection;

    public interface IPropertyAccessor
    {
        PropertyInfo PropertyInfo
        {
            get;
        }

        string Name
        {
            get;
        }

        object GetValue(object source);

        void SetValue(object source, object value);
    }


    public static class PropertyInfoHelper
    {
        private static readonly Dictionary<PropertyInfo, IPropertyAccessor> Cache 
            = new Dictionary<PropertyInfo, IPropertyAccessor>();
 
        public static IPropertyAccessor GetFastAccessor(PropertyInfo propertyInfo)
        {
            IPropertyAccessor result;
            lock (Cache)
            {
                if (!Cache.TryGetValue(propertyInfo, out result))
                {
                    result = CreateAccessor(propertyInfo);
                    Cache.Add(propertyInfo, result);
                }
            }
            return result;
        }

   
        public static IPropertyAccessor CreateAccessor(PropertyInfo propertyInfo)
        {
            return (IPropertyAccessor)Activator.CreateInstance(
                        typeof(PropertyWrapper<,>).MakeGenericType
                           (propertyInfo.DeclaringType, propertyInfo.PropertyType), propertyInfo);
        }
    }

    internal class PropertyWrapper<TObject, TValue> : IPropertyAccessor
    {
        private readonly PropertyInfo _propertyInfo;

        private readonly Func<TObject, TValue> _getMethod;
        private readonly Action<TObject, TValue> _setMethod;


        public PropertyWrapper(PropertyInfo propertyInfo)
        {
            _propertyInfo = propertyInfo;

            MethodInfo mGet = propertyInfo.GetGetMethod(true);
            MethodInfo mSet = propertyInfo.GetSetMethod(true);

            _getMethod = (Func<TObject, TValue>)Delegate.CreateDelegate
                    (typeof(Func<TObject, TValue>), mGet);
            _setMethod = (Action<TObject, TValue>)Delegate.CreateDelegate
                    (typeof(Action<TObject, TValue>), mSet);
        }

        object IPropertyAccessor.GetValue(object source)
        {
            return _getMethod((TObject)source);
        }
        void IPropertyAccessor.SetValue(object source, object value)
        {
            _setMethod((TObject)source, (TValue)value);
        }


        public string Name
        {
            get
            {
                return _propertyInfo.Name;
            }
        }

        public PropertyInfo PropertyInfo
        {
            get
            {
                return _propertyInfo;
            }
        }

    }

}