﻿namespace UsefulTools.ExtensionMethods
{
    using System;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Linq;
    using Core;

    public static class Collections
    {

        public static IEnumerable<TSource> SelectManyRecursive<TSource>(this IEnumerable<TSource> source,
            Func<TSource, IEnumerable<TSource>> selector, Func<TSource, bool> predicate = null, bool leafsOnly = false)
        {
            Throw.IfNull(source, "source");

            if (predicate == null)
            {
                predicate = x => true;
            }

            if (leafsOnly)
            {
                return source.Where(predicate).SelectMany(Functional.Y<TSource, IEnumerable<TSource>>
                    (f => obj =>
                    {
                        var children = selector(obj) ?? new List<TSource>();
                        var selected = children.Where(predicate).SelectMany(f);
                        return selected.Any() ? selected : selected.Prepend(obj);
                    }));
            }

            return source.Where(predicate).SelectMany(Functional.Y<TSource, IEnumerable<TSource>>
                (f => obj => selector(obj).Where(predicate).SelectMany(f).Prepend(obj)));

        }

        public static IEnumerable<string> AsStrings<T>(this IEnumerable<T> elems)
        {
            Throw.IfNull(elems);
            return elems.Select(e => e.ToString());
        }
        public static IList<Tuple<T1, T2>> Pairs<T1, T2>(params object[] elements)
        {
            Throw.IfNull(elements);
            Throw.If(elements.Any(x => x == null));
            Throw.If(elements.Length % 2 != 0);
            var result = new List<Tuple<T1, T2>>();
            for (int i = 0; i < elements.Length; i += 2)
            {
                result.Add(new Tuple<T1, T2>((T1)elements[i], (T2)elements[i + 1]));
            }
            return result;
        }
       
        public static ReadOnlyCollection<T> ToReadonly<T>(this IEnumerable<T> collection)
        {
            Throw.IfNull(collection);
            return new ReadOnlyCollection<T>(collection.ToList());
        }
        public static IDictionary<T1, T3> MapValues<T1, T2, T3>(this IDictionary<T1, T2> dict, Func<T1, T2, T3> selector)
        {
            Throw.IfNull(dict);
            return dict.Select(pair => new KeyValuePair<T1, T3>(pair.Key, selector(pair.Key, pair.Value)))
                .ToDictionary(pair => pair.Key, pair => pair.Value);
        }
        public static void AddRange<T>(this ICollection<T> collection, params T[] toAdd)
        {
            Throw.IfNull(collection);
            Throw.IfNull(toAdd);
            collection.AddRange(toAdd.AsEnumerable());
        }
        public static double AverageOrZero<T>(this IEnumerable<T> collection, Func<T, int> func)
        {
            Throw.IfNull(collection);
            return collection.Any() ? collection.Average(func) : 0;
        }
        public static double AverageOrZero<T>(this IEnumerable<T> collection, Func<T, long> func)
        {
            Throw.IfNull(collection);
            return collection.Any() ? collection.Average(func) : 0;
        }
        public static IEnumerable<T> WhereNot<T>(this IEnumerable<T> collection, Func<T, bool> func)
        {
            Throw.IfNull(collection);
            return collection.Where(e => !func(e));
        }
        public static IEnumerable<T> Flatten<T>(this IEnumerable<IEnumerable<T>> collection)
        {
            Throw.IfNull(collection);
            return collection.SelectMany(e => e);
        }
        public static void AddRange<T>(this ICollection<T> collection, IEnumerable<T> toAdd)
        {
            Throw.IfNull(collection);
            Throw.IfNull(toAdd);
            foreach (T item in toAdd)
            {
                collection.Add(item);
            }

        }
        public static Dictionary<TKey, T> ToSparseDictionary<T, TKey>(this IEnumerable<T> collection,
            ICollection<TKey> keySpace, Func<T, TKey> keySelector)
        {
            Throw.IfNull(collection);
            var list = new List<T>();
            foreach (T item in collection)
            {
                var key = keySelector(item);
                if (keySpace.Contains(key))
                {
                    keySpace.Remove(key);
                    list.Add(item);
                    if (keySpace.Count == 0)
                    {
                        break;
                    }
                }
            }
            return list.ToDictionary(keySelector);
        }
      
        public static IEnumerable<T> ToEmptyIfNull<T>(this IEnumerable<T> enumerable)
        {
            return enumerable ?? Enumerable.Empty<T>();
        }
        public static bool ContainsAny<T>(this IEnumerable<T> enumerable, T val1, T val2)
        {
            Throw.IfNull(enumerable);
            return enumerable.Any(obj => obj.Equals(val1) || obj.Equals(val2));
        }
        public static bool ContainsAny<T>(this IEnumerable<T> enumerable, T val1, T val2, T val3)
        {
            Throw.IfNull(enumerable);
            return enumerable.Any(obj => obj.Equals(val1) || obj.Equals(val2) || obj.Equals(val3));
        }

        public static NotifyingCollection<T> ToObsCollection<T>(this IEnumerable<T> collection)
        {
            return new NotifyingCollection<T>(collection);
        }
    }
}